package cn.com.qmth.examcloud.web.interceptor;

import cn.com.qmth.examcloud.commons.util.JsonUtil;
import cn.com.qmth.examcloud.commons.util.ThreadLocalUtil;
import cn.com.qmth.examcloud.web.enums.HttpServletRequestAttribute;
import cn.com.qmth.examcloud.web.support.*;
import com.google.common.collect.Maps;
import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.ThreadContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.HttpStatus;
import org.springframework.web.method.HandlerMethod;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.ModelAndView;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * 首发拦截器
 *
 * @author WANGWEI
 * @date 2018年10月22日
 * @Copyright (c) 2018-2020 WANGWEI [QQ:522080330] All Rights Reserved.
 */
public class FirstInterceptor implements HandlerInterceptor {

    private static final Logger LOG = LoggerFactory.getLogger(FirstInterceptor.class);

    @Override
    public boolean preHandle(HttpServletRequest request, HttpServletResponse response,
                             Object handler) throws Exception {
        ThreadLocalUtil.clearAll();
        String traceId = request.getHeader("Trace-Id");
        if (StringUtils.isBlank(traceId)) {
            traceId = ThreadLocalUtil.next();
        } else {
            ThreadLocalUtil.setTraceId(traceId);
        }

        // 设置log4j线程上下文
        ThreadContext.put("TRACE_ID", traceId);
        ThreadContext.put("CALLER", "$$$$$");

        String path = request.getServletPath();
        String method = request.getMethod();

        Set<String> headerNames = new HashSet<>(20);

        if (LOG.isDebugEnabled()) {
            Map<String, String> headers = Maps.newHashMap();
            Enumeration<String> e = request.getHeaderNames();
            while (e.hasMoreElements()) {
                String name = e.nextElement();
                String value = request.getHeader(name);
                headers.put(name, value);
                headerNames.add(name);
            }

            LOG.debug(String.format("[%s] - %s, headers = %s", method, path, JsonUtil.toJson(headers)));
        }

        if (path.equals("/error")) {
            response.setStatus(HttpStatus.NOT_FOUND.value());
            ServletUtil.returnJson(new StatusResponse(String.valueOf(HttpStatus.NOT_FOUND.value()), HttpStatus.NOT_FOUND.getReasonPhrase()), response);
            return false;
        }

        if ((!org.springframework.http.HttpMethod.OPTIONS.matches(method)) && handler instanceof HandlerMethod) {
            HandlerMethod handlerMethod = (HandlerMethod) handler;
            ApiId apiId = handlerMethod.getMethodAnnotation(ApiId.class);

            ApiInfo apiInfo = null;
            if (null != apiId) {
                apiInfo = ApiInfoHolder.getApiInfo(apiId.value());
            } else {
                apiInfo = ApiInfoHolder.getApiInfo(handlerMethod.getMethod());
            }

            String mapping = apiInfo.getMapping();
            request.setAttribute(HttpServletRequestAttribute.$_API_INFO.name(), apiInfo);

            LOG.debug("[mapping] ==> " + mapping);
            request.setAttribute(HttpServletRequestAttribute.$_MAPPING.name(), mapping);
        } else {
            String mapping = StringUtils.join("_[", path, "][", method, "]");

            LOG.debug("[mapping] --> " + mapping);
            request.setAttribute(HttpServletRequestAttribute.$_MAPPING.name(), mapping);
        }

        return true;
    }

    @Override
    public void postHandle(HttpServletRequest request, HttpServletResponse response, Object handler,
                           ModelAndView modelAndView) throws Exception {
    }

    @Override
    public void afterCompletion(HttpServletRequest request, HttpServletResponse response,
                                Object handler, Exception ex) throws Exception {
        // 清理log4j线程上下文
        ThreadContext.clearAll();
    }

}