package cn.com.qmth.examcloud.web.interceptor;

import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.collections.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.HttpStatus;
import org.springframework.web.method.HandlerMethod;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.ModelAndView;

import cn.com.qmth.examcloud.api.commons.security.bean.User;
import cn.com.qmth.examcloud.commons.util.ThreadLocalUtil;
import cn.com.qmth.examcloud.web.enums.HttpServletRequestAttribute;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import cn.com.qmth.examcloud.web.support.ServletUtil;
import cn.com.qmth.examcloud.web.support.StatusResponse;

/**
 * 顺序锁拦截器
 *
 * @author WANGWEI
 * @date 2019年2月22日
 * @Copyright (c) 2018-2020 WANGWEI [QQ:522080330] All Rights Reserved.
 */
public class SeqlockInterceptor implements HandlerInterceptor {

	private static final Logger LOG = LoggerFactory.getLogger(SeqlockInterceptor.class);

	/**
	 * redis client
	 */
	private RedisClient redisClient;

	private static final String GLOBAL_LOCK_PREFIX = "$_LOCK_G:";

	private static final String SESSION_LOCK_PREFIX = "$_LOCK_S:";

	private static final String LOCK_ATTRIBUTE = "$sequenceLock";

	/**
	 * 构造函数
	 *
	 * @param redisClient
	 */
	public SeqlockInterceptor(RedisClient redisClient) {
		super();
		this.redisClient = redisClient;
	}

	@Override
	public boolean preHandle(HttpServletRequest request, HttpServletResponse response,
			Object handler) throws Exception {
		if(LOG.isDebugEnabled()) {
			LOG.debug(String.format("[%s] - %s", request.getMethod(), request.getServletPath()));
		}

		if (handler instanceof HandlerMethod) {

			HandlerMethod handlerMethod = (HandlerMethod) handler;
			GlobalSequenceLock globalSeqlock = handlerMethod
					.getMethodAnnotation(GlobalSequenceLock.class);
			SessionSequenceLock sessionSeqlock = handlerMethod
					.getMethodAnnotation(SessionSequenceLock.class);

			String mapping = (String) request
					.getAttribute(HttpServletRequestAttribute.$_MAPPING.name());

			if (null != globalSeqlock) {
				String key = GLOBAL_LOCK_PREFIX + mapping;

				if (redisClient.setIfAbsent(key, ThreadLocalUtil.getTraceId(), 60 * 5)) {
					LOG.info("partition locked");

					request.setAttribute(LOCK_ATTRIBUTE, key);
					return true;
				} else {
					response.setStatus(HttpStatus.CONFLICT.value());
					ServletUtil.returnJson(new StatusResponse("409", "请求等待,请稍后重试!"), response);
					return false;
				}

			} else if (null != sessionSeqlock) {

				User user = (User) request
						.getAttribute(HttpServletRequestAttribute.$_ACCESS_USER.name());

				String key = SESSION_LOCK_PREFIX + user.getKey();

				if (redisClient.setIfAbsent(key, ThreadLocalUtil.getTraceId(), 60 * 5)) {
					LOG.info("session locked");

					request.setAttribute(LOCK_ATTRIBUTE, key);
					return true;
				} else {
					response.setStatus(HttpStatus.CONFLICT.value());
					ServletUtil.returnJson(new StatusResponse("409", "请稍后重试... ..."), response);
					return false;
				}
			}
		}

		return true;
	}

	@Override
	public void postHandle(HttpServletRequest request, HttpServletResponse response, Object handler,
			ModelAndView modelAndView) throws Exception {
	}

	@Override
	public void afterCompletion(HttpServletRequest request, HttpServletResponse response,
			Object handler, Exception ex) throws Exception {

		Object key = request.getAttribute(LOCK_ATTRIBUTE);
		if (null != key) {
			redisClient.delete((String) key);
		}

		@SuppressWarnings("unchecked")
		List<String> keyList = (List<String>) ServletUtil.getRequest()
				.getAttribute(HttpServletRequestAttribute.$_CUSTOM_SEQUENCE_LOCK.name());
		if (CollectionUtils.isNotEmpty(keyList)) {
			for (String cur : keyList) {
				redisClient.delete(cur);
			}
		}
	}

}
