package cn.com.qmth.examcloud.web.support;

import cn.com.qmth.examcloud.api.commons.exchange.BaseResponse;
import cn.com.qmth.examcloud.commons.util.JsonMapper;
import cn.com.qmth.examcloud.commons.util.ObjectUtil;
import cn.com.qmth.examcloud.commons.util.ThreadLocalUtil;
import cn.com.qmth.examcloud.web.config.LogProperties;
import cn.com.qmth.examcloud.web.enums.HttpServletRequestAttribute;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Pointcut;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Component;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.Serializable;
import java.util.Collection;
import java.util.Map;

/**
 * spring mvc controller aspect.
 *
 * @author WANGWEI
 * @date 2019年1月30日
 * @Copyright (c) 2018-2020 WANGWEI [QQ:522080330] All Rights Reserved.
 */
@Component
@Aspect
public class ControllerAspect {

    private static final Logger LOG = LoggerFactory.getLogger(ControllerAspect.class);

    @Autowired(required = false)
    HttpMethodProcessor httpMethodProcessor;

    @Autowired
    LogProperties logProperties;

    @Autowired(required = false)
    RedisClient redisClient;

    /**
     * 构造函数
     */
    public ControllerAspect() {
        super();
        LOG.info("ControllerAspect init...");
    }

    private static String[] excludeFields = new String[]{"password", ".*Password"};

    /**
     * handlerMethods
     *
     * @author WANGWEI
     */
    @Pointcut("@annotation(org.springframework.web.bind.annotation.RequestMapping) "
            + "|| @annotation(org.springframework.web.bind.annotation.GetMapping) "
            + "|| @annotation(org.springframework.web.bind.annotation.PostMapping)"
            + "|| @annotation(org.springframework.web.bind.annotation.DeleteMapping)"
            + "|| @annotation(org.springframework.web.bind.annotation.PutMapping)")
    public void handlerMethods() {

    }

    /**
     * 接口处理
     *
     * @param joinPoint
     * @return
     * @author WANG WEI
     */
    @Around("handlerMethods()")
    public Object doAroundWebRequests(ProceedingJoinPoint joinPoint) {
        // 获取request对象
        HttpServletRequest request = ServletUtil.getRequest();
        String path = request.getServletPath();
        try {
            if ("/".equals(request.getServletPath())) {
                return joinPoint.proceed();
            }
        } catch (Throwable e) {
            LOG.error(e.getMessage(), e);
            throw new RuntimeException(e);
        }

        long startTime = System.currentTimeMillis();
        String method = request.getMethod();
        Object[] args = joinPoint.getArgs();

        if (LOG.isInfoEnabled()) {
            StringBuilder params = new StringBuilder();

            // api method params
            if (ArrayUtils.isNotEmpty(args)) {
                params.append(" methodParams: ");
                for (int i = 0; i < args.length; i++) {
                    Object curArg = args[i];
                    String content = this.parseContent(curArg);
                    params.append(content);

                    if (i < args.length - 1) {
                        params.append(" | ");
                    }
                }
            }

            // api url params
            Map<String, String[]> paramMaps = request.getParameterMap();
            if (MapUtils.isNotEmpty(paramMaps)) {
                params.append(" urlParams: ");
                for (Map.Entry<String, String[]> entry : paramMaps.entrySet()) {
                    String name = entry.getKey();
                    String[] values = entry.getValue();
                    params.append("&").append(name).append("=");

                    if (ArrayUtils.isNotEmpty(values)) {
                        if (name.toLowerCase().indexOf(excludeFields[0]) >= 0) {
                            params.append("***");
                        } else {
                            params.append(values[0]);
                        }
                    }
                }
            }

            LOG.info(String.format("[request][%s] - %s %s", method, path, params.toString()));
        }

        Object ret;
        try {
            if (null != httpMethodProcessor) {
                httpMethodProcessor.beforeMethod(request, args);
            }

            // 执行
            ret = joinPoint.proceed();

            request.setAttribute(HttpServletRequestAttribute.$_EXCEPTION_HAPPENED.name(), new Boolean(false));
        } catch (Throwable e) {
            request.setAttribute(HttpServletRequestAttribute.$_EXCEPTION_HAPPENED.name(), new Boolean(true));

            if (null != httpMethodProcessor) {
                try {
                    httpMethodProcessor.onException(request, args, e);
                } catch (Exception ex) {
                    LOG.error("processor1 - " + ex.getMessage(), ex);
                }
            }

            LOG.error(String.format("[execute fail][%s] - %s, cost %sms, err is %s"
                    , method, path, System.currentTimeMillis() - startTime, e.getMessage()));
            throw new RuntimeException(e);
        }

        if (null != httpMethodProcessor) {
            try {
                httpMethodProcessor.onSuccess(request, args, ret);
            } catch (Exception ex) {
                LOG.error("processor2 - " + ex.getMessage(), ex);
            }
        }

        if (null != ret && BaseResponse.class.isAssignableFrom(ret.getClass())) {
            BaseResponse baseResponse = (BaseResponse) ret;
            baseResponse.setCost(System.currentTimeMillis() - startTime);
        }

        if (LOG.isDebugEnabled() && logProperties.isNormalResponseLogEnable()) {
            if (ret == null) {
                LOG.debug("status = ok, responseMsg = void");
            } else if (ret instanceof ResponseEntity) {
                ResponseEntity<?> re = (ResponseEntity<?>) ret;
                Object body = re.getBody();

                String content = this.parseContent(body);
                LOG.debug(String.format("status = %s, responseMsg = %s",
                        re.getStatusCodeValue(), content));
            } else {
                String content = this.parseContent(ret);
                LOG.debug(String.format("status = ok, responseMsg = %s", content));
            }
        }

        HttpServletResponse response = ServletUtil.getResponse();
        response.setHeader("Trace-Id", ThreadLocalUtil.getTraceId());
        response.setHeader("cost", String.valueOf(System.currentTimeMillis() - startTime));

        LOG.info(String.format("[response][%s] - %s, cost %sms", method, path, System.currentTimeMillis() - startTime));
        return ret;
    }

    private String parseContent(Object obj) {
        if (obj == null) {
            // content empty...
            return "";
        }

        if (ObjectUtil.isBaseDataType(obj.getClass()) || obj instanceof String) {
            return obj.toString();
        }

        boolean jsonEnable = false;
        if (obj instanceof Collection) {
            jsonEnable = true;
        } else if (obj instanceof Map) {
            jsonEnable = true;
        } else if (obj instanceof Serializable) {
            String classPath = obj.getClass().getName();

            if (classPath.startsWith("org.springframework.data.domain.PageImpl")
                    || classPath.startsWith("cn.com.qmth")
                    || classPath.startsWith("com.qmth")) {
                jsonEnable = true;
            }
        }

        if (!jsonEnable) {
            // content ignore...
            return "...";
        }

        String content = null;
        try {
            // content to json
            content = new JsonMapper().toJson(obj);
        } catch (Exception e) {
            LOG.warn("parseContent " + e.getMessage());
        }

        if (content == null) {
            return "......";
        }

        if (content.length() > logProperties.getResponseLogJsonMaxSize()) {
            // content too large...
            return content.substring(0, logProperties.getResponseLogJsonMaxSize()) + " [MORE...]";
        }

        return content;
    }

}
