package cn.com.qmth.examcloud.web.support;

import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.util.JsonMapper;
import cn.com.qmth.examcloud.web.config.SystemProperties;
import cn.com.qmth.examcloud.web.enums.HttpServletRequestAttribute;
import cn.com.qmth.examcloud.web.exception.ApiFlowLimitedException;
import cn.com.qmth.examcloud.web.exception.SequenceLockException;
import cn.com.qmth.examcloud.web.jpa.DataIntegrityViolationTransverter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.dao.DataIntegrityViolationException;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.validation.BindingResult;
import org.springframework.validation.ObjectError;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.MissingServletRequestParameterException;
import org.springframework.web.bind.annotation.ControllerAdvice;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.multipart.support.MissingServletRequestPartException;

import javax.servlet.http.HttpServletRequest;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import java.util.List;
import java.util.Set;

/**
 * 异常转换器
 *
 * @author WANGWEI
 * @date 2019年1月30日
 * @Copyright (c) 2018-2020 WANGWEI [QQ:522080330] All Rights Reserved.
 */
@ControllerAdvice
@ResponseBody
public class CustomExceptionHandler {

    /**
     * 接口日志
     */
    private static final Logger LOG = LoggerFactory.getLogger(CustomExceptionHandler.class);

    /**
     * 异常处理
     *
     * @param e
     * @param request
     * @return
     * @author WANGWEI
     */
    @ExceptionHandler(RuntimeException.class)
    public ResponseEntity<StatusResponse> handleException(RuntimeException e, HttpServletRequest request) {
        Throwable cause = e.getCause();
        StatusResponse body = null;

        if (null == cause) {
            body = new StatusResponse(SystemProperties.APP_CODE + "-500", "系统异常");
            cause = e;
        } else if (cause instanceof ApiFlowLimitedException) {
            body = new StatusResponse("503", "limited. RPC");
        } else if (cause instanceof StatusException) {
            StatusException se = (StatusException) cause;
            body = new StatusResponse(SystemProperties.APP_CODE + "-" + se.getCode(), se.getDesc());
        } else if (cause instanceof DataIntegrityViolationException) {
            try {
                DataIntegrityViolationTransverter.throwIfDuplicateEntry((DataIntegrityViolationException) cause);
                body = new StatusResponse(SystemProperties.APP_CODE + "-500", "数据完整性错误");
            } catch (StatusException se) {
                body = new StatusResponse(se.getCode(), se.getDesc());
            }
        } else if (cause instanceof IllegalStateException) {
            body = new StatusResponse(SystemProperties.APP_CODE + "-500", cause.getMessage());
        } else if (cause instanceof javax.validation.ConstraintViolationException) {
            javax.validation.ConstraintViolationException cvExcp = (ConstraintViolationException) cause;
            Set<ConstraintViolation<?>> constraintViolations = cvExcp.getConstraintViolations();

            StringBuffer errorMsg = new StringBuffer();
            boolean isFirst = true;
            for (ConstraintViolation<?> cv : constraintViolations) {
                if (isFirst) {
                    errorMsg.append(cv.getMessage());
                    isFirst = false;
                } else {
                    errorMsg.append("; ").append(cv.getMessage());
                }
            }
            body = new StatusResponse(SystemProperties.APP_CODE + "-500", errorMsg.toString());
        } else if (cause instanceof SequenceLockException) {
            body = new StatusResponse(SystemProperties.APP_CODE + "-500", cause.getMessage());
        } else if (cause instanceof org.springframework.jdbc.CannotGetJdbcConnectionException
                || cause instanceof org.springframework.orm.jpa.JpaSystemException
                || cause instanceof org.springframework.transaction.CannotCreateTransactionException) {
            body = new StatusResponse("503", "limited. JDBC");
        } else {
            body = new StatusResponse(SystemProperties.APP_CODE + "-500", "系统异常");
            cause = e;
        }

        return asResult(cause, body, request);
    }

    /**
     * 异常处理
     *
     * @param e
     * @param request
     * @return
     * @author WANGWEI
     */
    @ExceptionHandler(MethodArgumentNotValidException.class)
    public ResponseEntity<StatusResponse> handleException(MethodArgumentNotValidException e, HttpServletRequest request) {
        BindingResult result = e.getBindingResult();
        List<ObjectError> allErrors = result.getAllErrors();
        StringBuffer errorMsg = new StringBuffer();

        boolean isFirst = true;
        for (ObjectError err : allErrors) {
            if (isFirst) {
                errorMsg.append(err.getDefaultMessage());
                isFirst = false;
            } else {
                errorMsg.append("; ").append(err.getDefaultMessage());
            }
        }

        StatusResponse body = new StatusResponse(SystemProperties.APP_CODE + "-500", errorMsg.toString());
        return asResult(e, body, request);
    }

    /**
     * 异常处理
     *
     * @param e
     * @param request
     * @return
     * @author WANGWEI
     */
    @ExceptionHandler(MissingServletRequestParameterException.class)
    public ResponseEntity<StatusResponse> handleException(MissingServletRequestParameterException e, HttpServletRequest request) {
        StatusResponse body = new StatusResponse(SystemProperties.APP_CODE + "-500", e.getMessage());
        return asResult(e, body, request);
    }

    /**
     * 异常处理
     *
     * @param e
     * @param request
     * @return
     * @author WANGWEI
     */
    @ExceptionHandler(MissingServletRequestPartException.class)
    public ResponseEntity<StatusResponse> handleException(MissingServletRequestPartException e, HttpServletRequest request) {
        StatusResponse body = new StatusResponse(SystemProperties.APP_CODE + "-500", e.getMessage());
        return asResult(e, body, request);
    }

    /**
     * 异常处理
     *
     * @param e
     * @param request
     * @return
     * @author WANGWEI
     */
    @ExceptionHandler(HttpMessageNotReadableException.class)
    public ResponseEntity<StatusResponse> handleException(HttpMessageNotReadableException e, HttpServletRequest request) {
        StatusResponse body = new StatusResponse(SystemProperties.APP_CODE + "-500",
                "Required request body is missing");
        return asResult(e, body, request);
    }

    /**
     * 异常处理
     *
     * @param e
     * @param request
     * @return
     * @author WANGWEI
     */
    @ExceptionHandler(Exception.class)
    public ResponseEntity<StatusResponse> handleException(Exception e, HttpServletRequest request) {
        StatusResponse body = new StatusResponse(SystemProperties.APP_CODE + "-500", "系统错误");
        return asResult(e, body, request);
    }

    /**
     * 构建响应结果
     *
     * @param err
     * @param body
     * @param request
     * @return
     * @author WANGWEI
     */
    private ResponseEntity<StatusResponse> asResult(Throwable err, StatusResponse body, HttpServletRequest request) {
        boolean alwaysOK = alwaysOK(request);
        ApiInfo apiInfo = (ApiInfo) request.getAttribute(HttpServletRequestAttribute.$_API_INFO.name());

        boolean printStackTrace = true;
        if (null != apiInfo) {
            printStackTrace = !apiInfo.isWithoutStackTrace();
        }

        if (!printStackTrace) {
            String forcePrintStackTrace = System.getProperty("log.forcePrintStackTrace");
            if (null != forcePrintStackTrace && forcePrintStackTrace.equalsIgnoreCase(Boolean.toString(true))) {
                printStackTrace = true;
            }
        }

        HttpStatus httpStatus = null;
        if (alwaysOK) {
            if (err instanceof ApiFlowLimitedException
                    || err instanceof org.springframework.jdbc.CannotGetJdbcConnectionException
                    || err instanceof org.springframework.orm.jpa.JpaSystemException
                    || err instanceof org.springframework.transaction.CannotCreateTransactionException) {
                httpStatus = HttpStatus.SERVICE_UNAVAILABLE;
            } else {
                httpStatus = HttpStatus.OK;
            }
        } else {
            if (err instanceof ApiFlowLimitedException
                    || err instanceof org.springframework.jdbc.CannotGetJdbcConnectionException
                    || err instanceof org.springframework.orm.jpa.JpaSystemException
                    || err instanceof org.springframework.transaction.CannotCreateTransactionException) {
                httpStatus = HttpStatus.SERVICE_UNAVAILABLE;
            } else {
                httpStatus = HttpStatus.INTERNAL_SERVER_ERROR;
            }
        }

        String msg = String.format("[CustomExceptionHandler] status = %s, responseMsg = %s",
                httpStatus.value(), new JsonMapper().toJson(body));
        if (printStackTrace) {
            LOG.error(msg, err);
        } else {
            LOG.error(msg + ", err is " + err.getMessage());

            if (LOG.isDebugEnabled()) {
                LOG.debug(err.getMessage(), err);
            }
        }
        HttpHeaders headers = new HttpHeaders();
        headers.add("Content-Type","application/json;charset=utf-8");
        return new ResponseEntity<>(body, headers,httpStatus);
    }

    /**
     * 是否总是响应200
     *
     * @param request
     * @return
     * @author WANGWEI
     */
    private boolean alwaysOK(HttpServletRequest request) {
        boolean alwaysOK = false;
        Object attribute = request.getAttribute(HttpServletRequestAttribute.$_ALWAYS_OK.name());
        if (null != attribute) {
            try {
                alwaysOK = (boolean) attribute;
            } catch (Exception ex) {
                // ignore
            }
        }
        return alwaysOK;
    }

}